<!DOCTYPE html>

<html>
<head>
    <meta charset="utf-8"/>
    <title>Tauon Music Box - Audio Stream</title>
    <link href="radio/theme.css" rel="stylesheet" type="text/css">
</head>

<body>
<div class="container">

        <div id="img-holder">

            <img id='picture' src="radio/logo-bg.png" alt=""/> <br> <br>


        </div>

        <div id="meta">

            <div id="title-text">
            </div>
            <div id="artist-text">
            </div>

            <div id="seekback">
                <div id="seekfront">
                </div>
            </div>
            <div id='player'>
                <div class='button' id='play-button'>

                    <div id="play-icon"></div>
                    <div id='stop-icon'></div>
                </div>

                <div class="slidecontainer">
                    <input type="range" min="0" max="100" value="85" class="slider" id="vol-slider">
                </div>

            </div>

    </div>

    <div id="lyrics"></div>
</div>


<script src="radio/radio.js"></script>
</body>

</html>
